# -*- coding: utf-8 -*-
"""
Lumerical Data Handling
Created on Tue Aug 18 17:06:05 2020
@author: Vidar Flodgren
Github: https://github.com/DeltaMod
"""

#use this to set current directory without running code: os.chdir(os.path.dirname(sys.argv[0]))
import os
import sys
import time
import h5py
import matplotlib
import matplotlib as mpl
import tkinter as tk

from tkinter.filedialog import askopenfilename, askdirectory
from matplotlib import patches as ptc
from matplotlib import colormaps as cmaps
from matplotlib.transforms import Affine2D
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import axes3d #If you want to be able to use projection="3D", then you need this:
import scipy
import numpy as np
from scipy import integrate, interpolate, constants
import json
from collections import Counter
import natsort
import csv
import xlrd

#%% Importing and executing logging
import logging
try:
    from . custom_logger import get_custom_logger

    # Importing plot tools 
    from . utils_utils import *
    
except:
    from custom_logger import get_custom_logger
    # Importing plot tools 
    from utils_utils import *
    print("Loading plot_utils packages locally, since root folder is the package folder")

    logger = get_custom_logger("DMU_PLOTUTILS")
    
class CustomFormatter(logging.Formatter):

    magenta  = "\033[1;35m"
    lblue    = "\033[1;34m"
    yellow   = "\033[1;33m"
    red      = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset    = "\x1b[0m"
    format   = "%(asctime)s - %(name)s \n %(levelname)s: %(message)s (%(filename)s:%(lineno)d)"

    FORMATS = {
        logging.DEBUG: magenta + format + reset,
        logging.INFO: lblue + format + reset,
        logging.WARNING: yellow + format + reset,
        logging.ERROR: red + format + reset,
        logging.CRITICAL: bold_red + format + reset
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)

# create logger with 'spam_application'
logger = logging.getLogger("DMU_UTILS")
logger.setLevel(logging.DEBUG)

# create console handler with a higher log level
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)

ch.setFormatter(CustomFormatter())
if (logger.hasHandlers()):
    logger.handlers.clear()

logger.addHandler(ch)

"""
PLOT TOOLS
"""

def dummy_text_params(string,FIG,fontsize=12,usetex=True,visible=False,ezplot=True):
    #This function just guesses based on fontsize, there's no way to do this nicely so we won't try:
    
    dummy = {"width":None,"height":None} 
    HMOD = 2;  WMOD = 0.175;
    if ezplot:
        DPI = FIG.fig.dpi
    else:
        DPI = FIG.dpi
    dummy["height"] = (HMOD*fontsize)/DPI
    dummy["width"] = (WMOD * len(string) * fontsize )/DPI
       
    return(dummy)

    
#Getting Legend entries
def get_combined_legend(FIG):
    leg_handles = []
    leg_labels  = [] 
    for key in FIG.ax.keys():
        lh,ll = FIG.ax[key].get_legend_handles_labels()
        leg_labels.append(ll)
        leg_handles.append(lh)
    
    def flatten(xss):
        return [x for xs in xss for x in xs]
    return(flatten(leg_handles),flatten(leg_labels))

# Automatically adjust tick locations to intervals that do not require two decimal places
#%%
def adjust_ticks(ax,which="both",Nx=4,Ny=4,xpad=1,ypad=1,respect_zero =True,whole_numbers_only=False,powerlimits=(-2,3)):
    """
    Input: 
        ax: axis that we want to adjust ticks for, this should be done AFTER the axlims have been set.
        which: which ticks to adjust any of - ["both","xticks","yticks"]
        Nx: number of xticks, default = 4 (can be included even if which="yticks") 
        Ny: number of yticks, default = 4 (can be included even if which="xticks")
        xpad: symmetric overlap of extra xticks (so, if your range is [-0.5,0,0.5] then it will add [-1,0.5,0,0.5,1])
        ypad: symmetric overlap of extra yticks (so, if your range is [-0.5,0,0.5] then it will add [-1,0.5,0,0.5,1])        
        respect_zero: Forces the zero tick to be included
    Output: Adjusted axticks such that the number of ticks is as close to your chosen Nx and Ny as possible, without using than 2 sig fig e.g. 0.25 
    """
    
    def limcalc(lim,N,pad,respect_zero=True,whole_numbers_only=False):
        proxvals = np.array([0.01,0.02,0.05,0.1,0.2,0.5,1.0,2.0,5.0,10,20,50,100])
        rawspace = np.linspace(lim[0],lim[1],N)
        rawdiff  = np.diff(rawspace)[0]
        oom      = np.floor(np.log10(abs(rawdiff)))
        oomdiff  = rawdiff/10**oom    
        oomlim = lim/10**oom
        padlim = [oomlim[0]-pad*oomdiff,oomlim[1]+pad*oomdiff]
        prox = np.abs(proxvals - oomdiff)
        tickdiff = proxvals[np.where(prox == np.min(prox))[0]][0]
        
        if respect_zero == False:
            ticks = np.arange(round(padlim[0]/tickdiff)*tickdiff,round(padlim[1]/tickdiff)*tickdiff,tickdiff)
            
        elif respect_zero == True:
            ticks = np.concatenate([np.flip(np.arange(0,padlim[0],-tickdiff)[1:]) ,  np.arange(0,padlim[1],tickdiff)])
        
        
        if whole_numbers_only == True:
            
            tickooms = [np.min(np.floor(np.log10(np.abs([tick for tick in ticks if tick !=float(0)])))), np.max(np.floor(np.log10(np.abs([tick for tick in ticks if tick !=float(0)])))) ] 
          
            if tickooms[0] == tickooms[1]:
                newticks = []
                for tick in ticks:
                    if tickooms[0] == tickooms[1]:
                        numstr = str(float(f"{tick:.1f}"))
                        if numstr.split(".")[1] =="0" :
                            newticks.append(tick)
                        ticks = np.array(newticks)

        return(ticks*10**oom,[f"{val:.1e}" for val in ticks],oom)
        
   
    xfmt = ScalarFormatterForceFormat(); xfmt.set_powerlimits(powerlimits)
    yfmt = ScalarFormatterForceFormat(); yfmt.set_powerlimits(powerlimits)

    
    if which == "both" or which == "xticks":
        xlim = ax.get_xlim()
        xticks,xticklabels,xoom = limcalc(xlim,Nx,xpad,respect_zero,whole_numbers_only)
        ax.set_xticks(xticks)
        ax.set_xticklabels(xticklabels)
        if xoom >1:
            xfmt.set_format("%2d")
        ax.xaxis.set_major_formatter(xfmt)
        ax.set_xlim(xlim)
        
    if which == "both" or which == "yticks":
        ylim = ax.get_ylim()
        yticks,yticklabels,yoom = limcalc(ylim,Ny,ypad,respect_zero,whole_numbers_only)
        ax.set_yticks(yticks)
        ax.set_yticklabels(yticklabels)
        
        if yoom >1:
            yfmt.set_format("%2d")
        ax.yaxis.set_major_formatter(yfmt)
        ax.set_ylim(ylim)


def align_axis_zeros(axes):

    ylims_current = {}   #  Current ylims
    ylims_mod     = {}   #  Modified ylims
    deltas        = {}   #  ymax - ymin for ylims_current
    ratios        = {}   #  ratio of the zero point within deltas

    for ax in axes:
        ylims_current[ax] = list(ax.get_ylim())
                        # Need to convert a tuple to a list to manipulate elements.
        deltas[ax]        = ylims_current[ax][1] - ylims_current[ax][0]
        ratios[ax]        = -ylims_current[ax][0]/deltas[ax]
    
    for ax in axes:      # Loop through all axes to ensure each ax fits in others.
        ylims_mod[ax]     = [np.nan,np.nan]   # Construct a blank list
        ylims_mod[ax][1]  = max(deltas[ax] * (1-np.array(list(ratios.values()))))
                        # Choose the max value among (delta for ax)*(1-ratios),
                        # and apply it to ymax for ax
        ylims_mod[ax][0]  = min(-deltas[ax] * np.array(list(ratios.values())))
                        # Do the same for ymin
        ax.set_ylim(tuple(ylims_mod[ax]))        

class ScalarFormatterForceFormat(mpl.ticker.ScalarFormatter):
    """
    How to use:
        yformat = ScalarFormatterForceFormat()
        yformat.set_powerlimits((0,0))
        gca().yaxis.set_major_formatter(format)
    """
    def __init__(self, useOffset=True, useMathText=True):
        super().__init__(useOffset=useOffset, useMathText=useMathText)
        self.set_format()

    def set_format(self,form="%1.1f"):
        self.format = form  # Give format here
        
#%%
def ClearAxis(fig):
    for ax in fig.axes:
        ax.grid(which='major', visible = False)
        ax.grid(which='minor', visible = False)
        
def DefaultGrid(ax):
    ax.grid(which='major', color='darkgrey', linestyle='--')
    ax.grid(which='minor', color='#CCCCCC', linestyle=':')  



def Extract_Keithley_Labels(ddict):
    # === NW Labels === #
    
    if len(ddict["Settings"]["Operation Mode"]) == 2:
        NWID = "NW = " + ddict["emitter"]["NWID"]
        
    if len(ddict["Settings"]["Operation Mode"]) == 4:
        NWID = "Emitter: "+ddict["emitter"]["NWID"] + "  $\\rightarrow$  Receiver: "+ddict["detector"]["NWID"]
    
    # === Operation Label === #
    baseop = [label for label in ddict["Settings"]["Operation Mode"] if not any(substr in [label.lower()] for substr in ["common","bias"])][0]
    if "_" in ddict["Settings"]["Test Name"]:
        OpLabel = ddict["Settings"]["Test Name"].split("_")[1].split("#")[0] 
    else:
        if len(ddict["Settings"]["Operation Mode"]) == 1:
            OpLabel = [label for label in ddict["Settings"]["Operation Mode"] if "common" not in label.lower()][0]
            NWID = "NW = Unknown"
        if len(ddict["Settings"]["Operation Mode"]) == 2:
            OpLabel = [label for label in ddict["Settings"]["Operation Mode"] if "common" not in label.lower()][0]
        
        elif len(ddict["Settings"]["Operation Mode"]) == 4:
            OpLabel = [label for label in ddict["Settings"]["Operation Mode"] if not any(substr in [label.lower()] for substr in ["common","bias"])][0]
    OpLabel = "Operation: "+ OpLabel 
    try:
        Run     = ddict["LOG"]["RUN No."]
    except:
        Run     = "NoLog" 
        
    return({"OpLabel":OpLabel,"NWID":NWID,"baseOP":baseop,"Run":Run})
    
def Keithley_Plot_Tagger(ezfig, ddict,double=False):
    
    if double:
        mats = list(ddict.keys())
        ddict = ddict[mats[0]][list(ddict[mats[0]].keys())[0]]
        
    ax = ezfig.ax[0]
    #Line 0 File info
    line0 = "File Location: " + "Device Data\\" + ddict["Data directory"].split("Device Data\\")[1]
    
    #Line1: Device Info: Device Name (DFR1-GG-BR1) || NW ID (NW1//NW2) || Attempt to show Date of Data Plotted (2024-01-01)
    try:
        line1a = "Device: " + ddict["LOG"]["Device"]
    except: 
        line1a = "Device: " + "see filename"
    KeithDL = Extract_Keithley_Labels(ddict)
    line1b = KeithDL["NWID"]
    datestring = ddict["Data directory"].split("\\")[-2].replace("_","-").split("-")
    datestring = [sstr for sstr in datestring if all(ss.isnumeric() for ss in sstr)]
    line1c = "-".join(datestring)
    
    line1 = line1a+"    " + line1b + "    " + line1c
    #########
    #########
    #Line2: Type of Plot: Ideality // Voltage Sweep // Pulse Sweep // Pulse Ladder || Light Condition: Light on//Light Off
    line2a = KeithDL["OpLabel"]
    try:
        IDF = ezfig.IDF
        if type(ezfig.IDF) != None:
            line2a = "Operation: Ideality Fit"
    except:
        None
    try:
        line2b = "Light: " + str(bool(ddict["LOG"]["Light Microscope"]))
    except:
        line2b = "Light: " + "Unknown"
    line2c = "Run: "+ str(KeithDL["Run"])
    line2 = line2a+"    " + line2b +"    " + line2c

    
    #Annotating the Figure
    ax.annotate(line0, (0.5,0.97), xytext=None, xycoords='figure fraction', textcoords=None, arrowprops=None, annotation_clip=None, ha="center",fontsize=plt.rcParams["figure.titlesize"]*0.5)
    ax.annotate(line1, (0.5,0.94), xytext=None, xycoords='figure fraction', textcoords=None, arrowprops=None, annotation_clip=None, ha="center",fontsize=plt.rcParams["figure.titlesize"]*0.5)
    ax.annotate(line2, (0.5,0.91), xytext=None, xycoords='figure fraction', textcoords=None, arrowprops=None, annotation_clip=None, ha="center",fontsize=plt.rcParams["figure.titlesize"]*0.5)
    return(KeithDL)

#%%
class cmap_seq(object):
    """
    Goal of class: use for colourmapping profiles such that:
        - cmap types can be selected
        - The range of values before their full colour contents is selected
        - Alternatively the increment between each cmap value can be selected
        - Invoking the class allows the next value to be given (as to not rely on the for loop i)
        - Allowing for the class to be reset
    """
    def __init__(self):
        self.cmap = None
        self.i    = 0
        
        """
        kwargs:
            ==============
            cmap:  the name of the colormap you wish to use
            steps: the number of increments of the colormap between first and last values
            custom: enables custom mode where colour 1 and colour 2 (and the interp between them) can be seleced
            col1: colour 1
            col2: colour 2
            interp: value for interpolation speed (number of points between col1 and col2)
        """
        
    def set_cmap(self,**kwargs):
        kwargdict = {'cmap':'cmap','colormap':'cmap','colourmap':'cmap',
                     'steps':'steps','step':'steps','N':'steps',
                     'custom':'custom','cust':'custom',
                     'col1':'col1','color1':'col1','colour1':'col1',
                     'col2':'col2','color2':'col2','colour2':'col2',
                     'interp':'interp','interpolation':'interp'}
        kuniq = np.unique(list(kwargdict.keys()))
        kw = KwargEval(kwargs, kwargdict, cmap='viridis',steps=100,custom=False,col1=None,col2=None,interp=None)
        self.istep = 1/kw.steps
        self.cmap = mpl.colormaps(kw.cmap)
        self.col  = self.cmap(self.i)

        
    def iter_cmap(self):
        self.i += self.istep
        self.col  = self.cmap(self.i)
        
    def reset(self):
        self.i = 0
        
   
#%%  
def KwargEval(fkwargs,kwargdict,**kwargs):
    """
    A short function that handles kwarg assignment and definition using the same kwargdict as before. To preassign values in the kw.class, you use the kwargs at the end
    use: provide the kwargs fed to the function as fkwargs, then give a kwarg dictionary. 
    
    Example:
        
        kw = KwargEval(kwargs,kwargdict,pathtype='rel',co=True,data=None)
        does the same as what used to be in each individual function. Note that this function only has error handling when inputting the command within which this is called.
        Ideally, you'd use this for ALL kwarg assigments to cut down on work needed.
    """
    #create kwarg class 
    class kwclass:
        co = True
        pass
    
    #This part initialises the "default" values inside of kwclass using **kwargs. If you don't need any defaults, then you can ignore this.
    if len(kwargs)>0:
        for kwarg in kwargs:
            kval = kwargs.get(kwarg,False)
            try:
                setattr(kwclass,kwarg, kval)
                
            except:
                logger.warn(" ".join(['kwarg =',str(kwarg),'does not exist!',' Skipping kwarg eval.']))
    #Setting the class kwargs from the function kwargs!     
    for kwarg in fkwargs:
        fkwarg_key = kwarg
        if kwarg not in kwargdict.keys():
            kwarg_low = [key.lower() for key in kwargdict.keys()]
            if kwarg in kwarg_low:
                kidx = kwarg_low.index(kwarg)
                kwarg = list(kwargdict.keys())[kidx]
                
        kval = kwargs.get(kwarg,False)
        fkval = fkwargs.get(fkwarg_key,False)
        
        try:
            setattr(kwclass,kwargdict[kwarg], fkval)
            
        except:
            cprint(['kwarg =',kwarg,'does not exist!',' Skipping kwarg eval.'],mt = ['wrn','err','wrn','note'])
    return(kwclass)


#%%Finds all indices where the data set changes direction
def strictly_increasing(items,returns="all"):
    sublists = [[]]
    lid      = 0
    diffs = np.diff(items)
    
    if diffs[0]>=0:
        sublists[0].append(0)
        
    for i,diff in enumerate(diffs):
        if diff>= 0:
            sublists[lid].append(i+1)
        if diff<0:
            sublists.append([])
            lid += 1
    
    lengths = np.array([len(L) for L in sublists])
    maxlen = np.max(lengths)
        
    if maxlen<4:
        return(False)
    
    Longlocs =  np.where(lengths == np.max(lengths))[0]
    longest_list = sublists[int(Longlocs[0])]

    if longest_list[-1]<=len(diffs):
        longest_list = longest_list + list(np.arange(longest_list[-1]+1,len(diffs)+1))
    noise_indices = [i for i,a in enumerate(items) if i not in longest_list ]
    
    return({"sublists":sublists,"longest":longest_list,"noise":noise_indices})

def strictly_positive(items,returns="all"):
    sublists = [[]]
    lid      = 0
    diffs = items
    
        
    for i,diff in enumerate(diffs):
        if diff>= 0:
            sublists[lid].append(i)
        if diff<0:
            sublists.append([])
            lid += 1
    
    lengths = np.array([len(L) for L in sublists])
    maxlen = np.max(lengths)
        
    if maxlen<4:
        return(False)
    
    Longlocs =  np.where(lengths == np.max(lengths))[0]
    longest_list = sublists[int(Longlocs[0])]

    if longest_list[-1]<=len(diffs):
        longest_list = longest_list + list(np.arange(longest_list[-1],len(diffs)))
    noise_indices = [i for i,a in enumerate(items) if i not in longest_list ]
    
    return({"sublists":sublists,"longest":longest_list,"noise":noise_indices})

def find_turning_points(data):
    
    turning_points = [0]
    try:
        if data[1] > data[0]:
            increasing = True
        else:
            increasing = False
        step_size = abs(data[int(len(data)/4-2)] - data[int(len(data)/4-3)])
        sep_tp = []
        tp_i = 1
        for i in range(1, len(data)-1):
            dx = abs(data[i+1] - data[i])
            if data[i] > data[i+1] and dx > 0.5*step_size:
                if increasing == True:
                    turning_points.append(i)
                    sep_tp.append(turning_points[tp_i] - turning_points[tp_i-1])
                    tp_i += 1
                    increasing = False
                    
            elif data[i] < data[i+1] and dx > 0.1*step_size:
                if increasing == False:
                    turning_points.append(i)
                    sep_tp.append(turning_points[tp_i] - turning_points[tp_i-1])
                    tp_i += 1
                    increasing = True
                    
        if len(turning_points) == 1:
            turning_points = [0,len(data)]
        elif len(turning_points)>1:
            turning_points.append(len(data))
            
        if max(sep_tp) - min(sep_tp)>3:
            turning_points = [0,len(data)]
        return turning_points
    except:
        return turning_points


#%%Takes a list of lists, then returns them split up where the dependent variable changes direction. 

def segment_sweep(L,indices):
    swdat = []
    inds = indices
    slices = [[inds[i],inds[i+1]+1] for i in range(len(inds)-1)]
    
    slices[-1][1] = len(L)
    for slc in slices:
        swdat.append(L[slc[0]:slc[1]])
    return(swdat)


def get_tab20bc(output="list",grouping="pairs"):
    t20b = plt.get_cmap("tab20b")
    t20c = plt.get_cmap("tab20c")

    
    cmap20bc = []
    if grouping == "pairs":
        for i in [0,1,2,3]:
            for j in [0,2]:
                 cmap20bc.append(t20c(4*i + j))
    
        for i in [4,3,0,1,2]:
            for j in [0,2]:
                 cmap20bc.append(t20b(4*i + j))
                 
        for i in [4]:
            for j in [0,2]:
                 cmap20bc.append(t20c(4*i + j))
    elif grouping == "all":
        for i in [0,1,2,3,4]:
            for j in [0,1,2,3]:
                 cmap20bc.append(t20b(4*i + j))
                 
        for i in [0,1,2,3,4]:
            for j in [0,1,2,3]:
                 cmap20bc.append(t20c(4*i + j))
    cmap = mpl.colors.LinearSegmentedColormap.from_list("tab20bc", cmap20bc,N=len(cmap20bc))
    
    if output == "list":    
        return(cmap20bc)
    if output == "cmap":
        return(cmap)
    if output == "both":
        return(cmap20bc,cmap)
    
def get_rgbhex_color(color_name,ctype = "hex"):
    
    cmap = get_tab20bc(output="cmap",grouping="all")

    # Select a color from the colormap (normalized index between 0 and 1)
    colornames = ["dark violet", "violet", "light violet", "pale violet",
                  "dark lime", "lime", "light lime", "pale lime",
                  "dark tan", "tan", "light tan", "pale tan",
                  "dark red", "red", "light red", "pale red",
                  "dark lilac", "lilac", "light lilac", "pale lilac",
                  "dark blue", "blue", "light blue", "pale blue",
                  "dark orange", "orange", "light orange", "pale orange",
                  "dark green", "green", "light green", "pale green",
                  "dark lilac", "lilac", "light lilac", "pale lilac",
                  "dark grey", "grey", "light grey", "pale grey"]
    colordict = {colorname:colorID for colorID,colorname in enumerate(colornames)}
    if color_name not in colornames:
        raise ValueError(f"Invalid colour name: '{color_name}'\nValid names are: {', '.join(colornames)}")
        
    rgb = cmap(colordict[color_name]) 
    if ctype == "rgba":
        return(rgb)
   
    if ctype == "rgb":
        return(rgb[:3])
    
    if ctype == "rgb255":
        return(tuple([int(val*255) for val in rgb[:3]]))
    
    if ctype == "rgba255":
        return(tuple([int(val*255) for val in rgb]))    
        
    
    elif ctype == "hex":
        # Include alpha channel (opacity) if needed
        hex_color = '#{:02x}{:02x}{:02x}'.format(int(rgb[0] * 255), int(rgb[1] * 255), int(rgb[2] * 255))

        # If alpha exists, we handle it separately (adding a transparency value if RGBA)
        if len(rgb) > 3:
            # Normalize alpha (it should be between 0 and 1)
            alpha = int(rgb[3] * 255)
            hex_color += '{:02x}'.format(alpha)  # Add alpha value to the hex code

        return hex_color

class ScalarFormatterForceFormat(mpl.ticker.ScalarFormatter):
    def _set_format(self):  # Override function that finds format to use.
        self.format = "%1.1f"  # Give format here



# -*- coding: utf-8 -*-
"""
Created on Wed Sep 23 11:12:46 2020
@author: vidar
"""

import os
import glob,shutil
from setuptools import setup, find_packages
import ast

import distutils.cmd

try:
   from setupext_janitor import janitor
   CleanCommand = janitor.CleanCommand
except ImportError:
   CleanCommand = None

cmd_classes = {}
if CleanCommand is not None:
   cmd_classes['clean'] = CleanCommand
 
TrgtScr = "DMU//utils.py"


class CleanCommand(distutils.cmd.Command):
    """
    Our custom command to clean out junk files.
    """
    description = "Cleans out junk files we don't want in the repo"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        cmd_list = dict(
            DS_Store="find . -name .DS_Store -print0 | xargs -0 rm -f;",
            pyc="find . -name '*.pyc' -exec rm -rf {} \;",
            empty_dirs="find ./pages/ -type d -empty -delete;"
        )
        for key, cmd in cmd_list.items():
            os.system(cmd)



"""
Run commands to find modules from the given script file
"""
modules = set()
def visit_Import(node):
    for name in node.names:
        modules.add(name.name.split(".")[0])

def visit_ImportFrom(node):
    # if node.module is missing it's a "from . import ..." statement
    # if level > 0 it's a "from .submodule import ..." statement
    if node.module is not None and node.level == 0:
        modules.add(node.module.split(".")[0])
    
def module_filter(modules):
    mod2 = list(modules)
    for module in mod2:
         if module in ["mpl_toolkits","json","time","sys","os","tkinter","collections","csv","custom_logger","plot_utils","utils_utils","pickle"]:
             modules.remove(module)

    return(list(modules))
        


node_iter = ast.NodeVisitor()
node_iter.visit_Import = visit_Import
node_iter.visit_ImportFrom = visit_ImportFrom


with open(TrgtScr) as f:
    node_iter.visit(ast.parse(f.read()))


with open("README.md", "r") as fh:
    long_description = fh.read()


installREQ = [module_filter(modules)+["docutils>=0.3"]]
installREQ = [[item for item in installREQ[0] if item not in ["logging"]]]

setup(
    name="DMU",
    version="0.3.9",
    packages=find_packages(),
    scripts=[TrgtScr],

    # Project uses reStructuredText, so ensure that the docutils get
    # installed or upgraded on the target machine
    install_requires =installREQ,
    package_data={
        # If any package contains *.txt or *.rst files, include them:
        "": ["*.txt", "*.json"],
        # And include any *.msg files found in the "hello" package, too:
        "hello": ["*.msg"],
    }, 
    # metadata to display on PyPI
    author="Atli Vidar Már FLodgren",
    author_email="vidar.flodgren@sljus.lu.se",
    description="This package is used to store commonly used functions on pip for the sake of easy pulling for other code.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="hello world example examples",
    url="https://github.com/DeltaMod/DMU",   # project home page, if any
    #project_urls={
    #    "Bug Tracker": "https://bugs.example.com/HelloWorld/",
    #    "Documentation": "https://docs.example.com/HelloWorld/",
    #    "Source Code": "https://code.example.com/HelloWorld/",
    #},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Development Status :: 2 - Pre-Alpha"]

    # could also include long_description, download_url, etc.
)



import os
#%% Importing and executing logging
import logging

#%% Importing and executing logging
import logging
import logging
try:
    from . custom_logger import get_custom_logger
    logger = get_custom_logger("DMU_SEMUTILS")
    # Importing plot tools 
    from . utils_utils import *
    from . plot_utils import *
    
except:
    from custom_logger import get_custom_logger
    logger = get_custom_logger("DMU_SEMUTILS")
    # Importing plot tools 
    from utils_utils import *
    from plot_utils import *
    print("Loading plot_utils packages locally, since root folder is the package folder")

    
import tifffile
import svgwrite
from PIL import Image,ImageEnhance,ImageOps
import base64
from io import BytesIO
import numpy as np 
import os 
import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import subprocess

def ANY_Image_Enhance(PIL_im,brightness=None,contrast=None,sharpness=None,expand_range=True):
    PIL_im = PIL_im.convert("L")
    
    if expand_range:
        pixvals = np.array(PIL_im)
        
        
        pixvals = ((pixvals - pixvals.min()) / (pixvals.max()-pixvals.min())) * 255
        PIL_im = Image.fromarray(pixvals.astype(np.uint8))
    
    if brightness:
        enhancer = ImageEnhance.Brightness(PIL_im)
        PIL_im = enhancer.enhance(brightness)
        
    if contrast:
        enhancer = ImageEnhance.Contrast(PIL_im)
        PIL_im = enhancer.enhance(contrast)
        
    if sharpness:
        enhancer = ImageEnhance.Sharpness(PIL_im)
        PIL_im = enhancer.enhance(sharpness)
    
    
    return(PIL_im)

def SEM_Annotation_Finder(image_path):
    """
    Provide image path, and you shall be given the crop matrix used in IMAGE.crop([left,upper,right,lower]). 
    ----------
    image_path : r'path\to\file.tif'
        

    crop_matrix = [left,upper,right,lower]
    
    If no annotation is found, then this crop matrix 

    """
    # Load image and convert to grayscale
    im = Image.open(image_path).convert('L')  # Convert to grayscale
 
    # Convert image to a numpy array
    im_array = np.array(im)
 
    # Apply a binary threshold (0 = black, 255 = white)
    binary_image = np.where(im_array == 0, 0, 255).astype(np.uint8)
 
    for row_idx, row in enumerate(binary_image):
        black_pixel_count = np.sum(row == 0)
        if black_pixel_count > 0.9 * len(row):  # 90% of the row is black
            l_crop = im.height - row_idx
            if l_crop%2!=0:
                l_crop+=1
            imcrop = [0,0,0,im.height - row_idx]    
            return(imcrop)
    
    imcrop = [0,0,0,0]
    return(imcrop)

          
def split_crop_bounds_evenly(length,newlength,offset=0):
    crop = length-newlength
    if crop%2==0:
        cropl=cropr=crop/2
    else:
        cropl=(crop-1)/2
        cropr = crop-cropl
    return(cropl+offset,length-cropr+offset)

def SEM_Scalebar_Generator(image_path, svg_output, scalebar_style = {},txt_style={}, imcrop=[0,0,0,0], savefile=True, resize=None, 
                           remove_annotation=True, resampling="bicubic",force_aspect=False,delta_offset=[0,0],rotation=0,crop_rescale=True,tweak_aspect=[1,1],filterdict={},recalculate_stroke_width=True):
    """
    Example image_path = 'DFR1-HE_BR204.tif' (or any literal string address)
    Example svg_output = 'output.svg'
    If svg_output is set to "Auto" then the name image_path+anno.svg is used
    
    scalebar_style and txt_style refer to the possible parameters we pass in a dict. Below are two examples that contain something in EVERY field possible currently.
    
    scalebar_style = {"frame":True,
                      "framepad":[30,2],
                      "stroke_width":4, %this should be calculated from a percentage of the image width
                      "stroke_style":"line",
                      "bar_color":"white",
                      "frame_color":"black",
                      "frame_opacity":0.6,
                      "location":"lower right",
                      "location_padding":[0.03,0.05],
                      "bar_ratio":[1/6,1/40]}
    
    txt_style={"font":"Arial",
               "fontsize":"Auto",
               "font_weight":"normal",
               "font_style":"normal",
               "text_decoration":"none",
               "color":"white"}
    
    Example use 1:
    dwg = SEM_Scalebar_Generator(image_path, svg_output, scalebar_style=scalebar_style,txt_style=txt_style, imcrop=[0,0,0,0])
    
    Example use 2:
        for file in [f for f in os.listdir() if f.endswith(".tif")]:
        dwg = SEM_Scalebar_Generator(image_path, "Auto", scalebar_style=scalebar_style,txt_style=txt_style, imcrop=[0,0,0,0])
    """
    
    """
    Defining the default scalebar and text settings
    """
    
    with tifffile.TiffFile(image_path) as tif:
        sem_metadata = tif.sem_metadata
        if sem_metadata == None:
            
            try:
                if tif.fei_metadata["System"]["SystemType"] == 'Nova NanoLab':                
                    sem_metadata = tif.fei_metadata["EScan"]
                    sem_metadata["sv_serial_number"] = ["Serial Code",tif.fei_metadata["System"]["SystemType"]]
                    sem_metadata["ap_image_pixel_size"] = ["ap_image_pixel_size",sem_metadata["PixelWidth"],"m"]
                    pix_size_string = "ap_image_pixel_size"
                    sem_metadata["ap_stage_at_t"] = ["rotation",tif.fei_metadata["Stage"]["SpecTilt"]]
                
            except:
                print("SEM MODEL NOT IMPLEMENTED!!! FIX IMPORTER")
        elif "SUPRA 35-29-41" in sem_metadata['sv_serial_number'][1]:
            pix_size_string = "ap_image_pixel_size"
        elif "Gemini" in sem_metadata['sv_serial_number'][1]:
            pix_size_string = "ap_image_pixel_size"
        elif "1560-95-96" in sem_metadata["sv_serial_number"][1]:
            pix_size_string = "ap_width"
        else:
            pix_size_string = "ap_pixel_size"
            
            
    
    def scalebar_style_dictgen(frame=None,framepad=[2,2],stroke_width=4,stroke_style="line",
                                 bar_color="white",frame_color="black",frame_opacity=1.0,
                                 location="lower left",location_padding=[0.03,0.05],
                                 bar_ratio=[1/6,1/40]):
            
        
        return({"frame":frame,"framepad":framepad,"stroke_width":stroke_width,"stroke_style":stroke_style,"bar_color":bar_color,
                "frame_color":frame_color,"location":location,"bar_ratio":bar_ratio,"location_padding":location_padding,"frame_opacity":frame_opacity})
    
    def text_style_dictgen(im,font_family="Arial",fontsize="Auto",font_fraction=1/20,font_weight="normal",font_style="normal",text_decoration="none",color="black"):
        if fontsize == "Auto":
            fontsize = int(im.height/20)
        if fontsize == "fraction":
            fontsize= int(im.height * font_fraction)
        
        return({"font_family":font_family,"fontsize":fontsize,"font_weight":font_weight,"font_style":font_style,"text_decoration":text_decoration,"color":color})
    
    
    resampling = {"nearest":Image.Resampling.NEAREST,"bicubic":Image.Resampling.BICUBIC,"bilinear":Image.Resampling.BILINEAR,"lancoz":Image.Resampling.LANCZOS,"box":Image.Resampling.BOX}[resampling.lower()]
    if svg_output == "Auto":
        svg_output = image_path.split(".")[0]+"_anno.svg"
        
    with Image.open(image_path) as im:
        buffer = BytesIO()
        imformat = im.format            
        
        if remove_annotation == True:
            annocrop= SEM_Annotation_Finder(image_path)        
            im = im.crop((annocrop[0],annocrop[1],im.width-annocrop[2],im.height-annocrop[3]))
            
        
        og_w,og_h = (im.width, im.height)
        
        if tweak_aspect != [1,1]:
            im = im.resize((int(im.width*tweak_aspect[0]),int(im.height*tweak_aspect[1])),resample=resampling)    
            og_w,og_h = (im.width, im.height)
        
        if force_aspect!=False:
            og_w,og_h = find_nearest_aspect_dim(im.width,im.height,force_aspect)
        
        if imcrop != [0,0,0,0]:
            im = im.crop((imcrop[0],imcrop[1],im.width-imcrop[2],im.height-imcrop[3]))
           
            
        if rotation != 0:
            
            im = im.rotate(rotation, expand=True)
            
            
        if force_aspect != False:
            xd,yd = find_nearest_aspect_dim(im.width,im.height,force_aspect)
            xcrop = split_crop_bounds_evenly(im.width, xd,offset=delta_offset[0])
            ycrop = split_crop_bounds_evenly(im.height, yd,offset=delta_offset[1])
            im = im.crop((xcrop[0],ycrop[0],xcrop[1],ycrop[1]))
        
        
            
        if crop_rescale == True:
            pix_rescale = 1/(og_w/im.width)
            im = im.resize((og_w,og_h),resample=resampling)
        else:
            pix_rescale = 1
        

            
        if resize != None:
            rszm = resize
            im = im.resize((int(im.width*rszm),int(im.height*rszm)),resample=resampling)
            pix_rescale *=1/(rszm)
        else:
            rszm = 1
            
        if len(filterdict.keys())!=0:
            im = ANY_Image_Enhance(im,**filterdict)
        
        im.format = imformat

        im.save(buffer, format=im.format)
        #im.save(buffer, format=im.format)  # Save the image in its original format to a buffer
        img_str = base64.b64encode(buffer.getvalue()).decode('utf-8')  # Encode image as base64
    
    
    txt  = text_style_dictgen(im,**txt_style)    

    
    
    sbar = scalebar_style_dictgen(**scalebar_style)
    if recalculate_stroke_width:
        sbar["stroke_width"] *=im.width/1000 
    
    txt_orig = txt.copy()
    sbar_orig = sbar.copy()
    
    OOM = {"nm":1e-9,"um":1e-6,"µm":1e-6,"mm":1e-3,"m":1e+0}
    #Find the scale parameters from the image in question: 
    
    if pix_size_string != "ap_width":
        pix_size = sem_metadata[pix_size_string][1] * OOM[sem_metadata[pix_size_string][2]]*pix_rescale
    elif pix_size_string == "ap_width":
        pix_size = sem_metadata[pix_size_string][1] * OOM[sem_metadata[pix_size_string][2]]*pix_rescale/og_w
        
    rtilt = np.radians(sem_metadata['ap_stage_at_t'][1])
    rrot = np.radians(rotation)
    
    if abs(sem_metadata['ap_stage_at_t'][1]) >= 5 and abs(rotation) > 10:
            L = pix_size/np.sin(rtilt)
            pix_size = L*np.cos(np.arcsin(np.cos(rrot)*np.sin(rtilt))) 
            #pix_size = np.sqrt(L**2*(np.cos(rrot)**2*np.cos(rtilt)**2 + np.sin(rrot)**2)) #L**2 - L**2*np.cos(rtilt)**2*np.sin(rrot)**2

    #Define bar length and height. Note that setting the height to zero removes it

    sbar["bar_length_target"] = int(im.width*sbar["bar_ratio"][0]) * pix_size 
    sbar["bar_height"] = int(im.height*sbar["bar_ratio"][1]) 
    
    def find_nearest_scale_bar(target_length):
        """
        We list all multipliers, powers of 10 we will consider, and calculate allowed values.
        """
        # List of allowed multipliers
        multipliers = np.array([1, 2, 5, 10, 20, 50, 100, 200, 500])
        scale_dict = {"1e+09":"Gm","1e+06":"Mm","1e+03":"km","1e+00":"m","1e-03":"mm","1e-06":"µm","1e-09":"nm"}
        OOMS = np.array([1e+9, 1e+6, 1e+3, 1e+0, 1e-3, 1e-6, 1e-9])    
        allowed_values = (multipliers[:, None] * OOMS).flatten()
        powers_array = np.tile(OOMS, len(multipliers))
        
        if target_length == 0:
            return 0, 1  # Return 0 for scale bar and 1 as power of 10 for simplicity
        
        # Find the index of the nearest value in allowed_values
        nearest_index = np.abs(allowed_values - target_length).argmin()
        
        # Get the nearest scale bar and its corresponding power of ten
        nearest_scale_bar = allowed_values[nearest_index]
        
        nearest_power_of_ten = powers_array[nearest_index]
        
        len_string = str(np.round(nearest_scale_bar/nearest_power_of_ten,1))
        if len_string.endswith(".0"):
            len_string = len_string.replace(".0","")
            
        len_string +=scale_dict["{:.0e}".format(nearest_power_of_ten)]

        return(nearest_scale_bar, nearest_power_of_ten, len_string)
    
    
    
    def scalebar_generation(svg_output,im,scalebar_style,txt_style,pix_size):
        #Draw direction and location determination
        sbar = scalebar_style; txt = txt_style

        best_scale, best_oom, len_string = find_nearest_scale_bar(sbar["bar_length_target"])
        
        bar_length = best_scale/pix_size
        
        
        if "lower" in sbar["location"]:
            y_coord = int(im.height - im.height*sbar["location_padding"][1])
            draw_dir_y = -1
        elif "upper" in sbar["location"]:
            y_coord = int(im.height*sbar["location_padding"][1])
            draw_dir_y = 1
        else:
            y_coord = int(im.height/2)
        if "left" in sbar["location"]:
            draw_dir_x = 1
            x_coord = int(im.width*sbar["location_padding"][0])
        elif "right" in sbar["location"]: 
            x_coord = int(im.width-im.width*sbar["location_padding"][0])
            draw_dir_x = -1
        else:
            x_coord = int(im.width/2)
            draw_dir_x = 0
        draw_dir = [draw_dir_x,draw_dir_y]
        location = [x_coord,y_coord]
        bar_middle_x = draw_dir[0] * bar_length/2 + location[0]
        bar_start = tuple(location)
        bar_end   = tuple([draw_dir[0] * bar_length + location[0],location[1]] )
        
        textloc = tuple([bar_middle_x,bar_start[1] + int(sbar["bar_height"]*1.2) * draw_dir[1]])
        
        # Create an SVG drawing with svgwrite
        dwg = svgwrite.Drawing(svg_output, profile='full', size=(im.width, im.height))
        
        # Add the raster image as base64 inside the SVG
        dwg.add(dwg.image(href=f"data:image/{im.format.lower()};base64,{img_str}",
                          insert=(0, 0), size=(im.width, im.height)))
        
        if sbar["frame"] != None:
            y_min = int(textloc[1] - txt["fontsize"]/1.5   - sbar["framepad"][1] - sbar["stroke_width"]*2)
            y_max = int(location[1] + sbar["bar_height"] + sbar["framepad"][1]+sbar["stroke_width"]*2)
            box_dxy = (np.abs(bar_start[0] - bar_end[0])+ sbar["framepad"][0] + sbar["stroke_width"]*2,y_max-y_min + sbar["framepad"][1])
            box_xy = (bar_middle_x - box_dxy[0]/2,np.mean([y_max,y_min])-box_dxy[1]/2) 
            
            dwg.add(dwg.rect(
                            insert=box_xy,  # Bottom-left corner of the rectangle
                            size=box_dxy,  # Width and height of the rectangle
                            fill=sbar["frame_color"],  # Fill color of the background (you can choose any color)
                            stroke="none",  # Optional stroke for the rectangle
                            stroke_width=0,
                            fill_opacity=sbar["frame_opacity"]
                        ))
        
        # Draw a vector line on top of the image
        dwg.add(dwg.line(start=bar_start, end=bar_end, 
                         stroke=sbar["bar_color"], stroke_width=sbar["stroke_width"]))
        
        if sbar["bar_height"] != 0:
            dwg.add(dwg.line(start=(bar_start[0],bar_start[1] + sbar["bar_height"]), end=(bar_start[0],bar_start[1] - sbar["bar_height"]), 
                             stroke=sbar["bar_color"], stroke_width=sbar["stroke_width"]))
            dwg.add(dwg.line(start=(bar_end[0],bar_end[1] + sbar["bar_height"]), end=(bar_end[0],bar_end[1] - sbar["bar_height"]), 
                             stroke=sbar["bar_color"], stroke_width=sbar["stroke_width"]))
        
        dwg.add(dwg.text(len_string,insert=textloc,text_anchor="middle", font_size=txt["fontsize"], font_family=txt["font_family"], fill=txt["color"]))

        # Save the SVG
    
        return(dwg)
    dwg = scalebar_generation(svg_output,im,sbar,txt,pix_size)
    if savefile == True:
        dwg.save()
    return({"svg":dwg,"im":im,"sbar":sbar_orig,"txt":txt_orig})
    
#%%
def find_nearest_aspect_dim(width, height, ratio):
    if width > height*ratio:
        width = int(height*ratio + 0.5)
    else:
        height = int(width/ratio + 0.5)
    return (width, height)



def SEM_Create_Insert(image_overview, inserts, filename="Auto",path="", scalebar_style = {},txt_style={},force_aspect=4/3,filterdict={},imcrop=[0,0,0,0],rotation=0):
    """
    This function wraps the SEM_Scalebar_Generator and creates inserts based on the provided information in the dicts. 
    The list of dicts must contain {"path":path\to\image,"size":1/N,"framecolor":CMAPformat, "framing":[x,y,w,h],"location":[1,1],"loc_type":"grid"}
    If you provide a size 1/N image, then the image_overview is scaled by Nx, and the number of choices you get in the grid mode becomes (N-1)
    In short, you get a 3x3 grid with a 1/2 size image, and a 7x7 grid from a 1/4 size, and a 15x15 from a 1/8 size and the coordinates for each becomes exactly as you expect.
    If you choose "loc_type":"pixels" then you simply provide the centre pixel that you want
    If you choose "figure_fraction" then you use fractions. These methods do not account for overlap.
    
    Parameters
    ----------
    image_overview : TYPE
        DESCRIPTION.
    inserts_dict : TYPE
        DESCRIPTION.
    scalebar_style : TYPE, optional
        DESCRIPTION. The default is {}.
    txt_style : TYPE, optional
        DESCRIPTION. The default is {}.

    Returns
    -------
    None.

    """
    def inserts_dict_generator(path=None,size=1/2,frame_color=plt.get_cmap("tab20c")(5), stroke_width=4, framing=[0,0,100,100],location=[1,1],loc_type="grid",delta_offset=[0,0],force_aspect=4/3,imcrop=[0,0,0,0],filterdict={},rotation=0):
        return({"path":path,"size":size,"frame_color":frame_color, "stroke_width":stroke_width, "framing":framing,"location":location,"loc_type":loc_type,"delta_offset":delta_offset,"imcrop":imcrop, "force_aspect":force_aspect,"filterdict":filterdict,"rotation":rotation})
    
    inserts = [inserts_dict_generator(**insert) for insert in inserts] #refactor inserts to adhere to formatting. Can't really be automatic, but will help I think
    
    with Image.open(image_overview) as im:
        width = im.width
        smallwidth = im.width*np.min([insert["size"] for insert in inserts])
        
    insert_widths  = []
    scalefactors   = []
    for insert in inserts:
        with Image.open(insert["path"]) as im:
            insert_widths.append(im.width)
            insert["base_scale_factor"] = smallwidth/im.width
            scalefactors.append(smallwidth/im.width)
    
    if np.min(scalefactors) < 1:
        #Scale up overview, don't scale anything that has same scalefactor. Scale up things with greater scale factors
        overview_scale = 1/np.min(scalefactors)
        for insert in inserts:
            insert["scale_factor"] = np.min(scalefactors) / insert["base_scale_factor"]  
        
    
    
    if np.min(scalefactors) >= 1:
        #Scale up everything else, don't scale anything that has same scalefactor. Scale up things with greater scale factors
        overview_scale = 1
        for insert in inserts:
            insert["scale_factor"] = insert["base_scale_factor"]
            
    if filename == "Auto":
        image_overview_name = image_overview.split("\\")[-1].split(".")[0]+"_combined_inserts.svg"
        image_overview_path = os.path.join(path,image_overview_name)
    
    else:
        image_overview_name = filename.split(".")[0] + ".svg"
        image_overview_path = os.path.join(path,image_overview_name)
    
    
    dwg = SEM_Scalebar_Generator(image_overview, image_overview_path, scalebar_style=scalebar_style,txt_style=txt_style, imcrop=imcrop,force_aspect=force_aspect,delta_offset=[0,0],resize=overview_scale,filterdict=filterdict,rotation=rotation)
    
    
    scalebar_style = dwg["sbar"]
    txt_style= dwg["txt"]
    insert_svglist = []
    for insert in inserts:
        s_factor     = insert["scale_factor"] 
        #location grid for the insert's sizing 
        points = np.linspace(0, 1, int((1/insert["size"]*2-1)**2))
        # Create the grid using np.meshgrid
        x_grid, y_grid = np.meshgrid(points*dwg["im"].width, points*dwg["im"].height)
        # Stack the x_grid and y_grid to create an array of tuples (x, y)
        cgrid = np.dstack((x_grid, y_grid)) # this is accessed as "row,col" 
        
        isbs = scalebar_style
     
        isbs["framepad"]         = [scalebar_style["framepad"][0],scalebar_style["framepad"][1]] 
        isbs["stroke_width"]     = scalebar_style["stroke_width"]
        isbs["bar_ratio"]        = [scalebar_style["bar_ratio"][0]/insert["size"]*0.8,scalebar_style["bar_ratio"][1]/insert["size"]]
        isbs["location_padding"] = [scalebar_style["location_padding"][0]/insert["size"],scalebar_style["location_padding"][1]/insert["size"]] 
        
        itxt = txt_style 
        if itxt["fontsize"] == "Auto":
            itxt["fontsize"] = int(dwg["im"].height/10)
        insert_path = insert["path"]
       
        insert_name = filename.split(".")[0]+"_"+insert_path.split("\\")[-1].split(".")[0]+"_insert.svg"
        insert_newpath = os.path.join(path,insert_name)
        insert_svg  = SEM_Scalebar_Generator(insert_path, insert_newpath, scalebar_style=isbs,txt_style=itxt, imcrop=insert["imcrop"],
                                                           force_aspect=insert["force_aspect"],delta_offset=insert["delta_offset"],resize=insert["scale_factor"],filterdict=insert["filterdict"],rotation=insert["rotation"],recalculate_stroke_width=False)
        
        insert_svglist.append(insert_svg)
        img_loc = cgrid[insert["location"][0],insert["location"][1]]
        img_loc = (img_loc[0],img_loc[1])
        if img_loc[0] >= dwg["im"].width/2:
            ilx=-isbs["stroke_width"]
        elif img_loc[0] < dwg["im"].width/2:
            ilx=isbs["stroke_width"]
        else:
            ilx = 0
            
        if img_loc[1] >= dwg["im"].height/2:
            ily=-isbs["stroke_width"]
        elif img_loc[1] < dwg["im"].height/2:
            ily=isbs["stroke_width"]
        else:
            ily = 0
            
        img_loc = (img_loc[0] + ilx, img_loc[1]+ily)
        
        href = 'file:///' + insert_newpath.replace('\\', '/')
        image = dwg["svg"].image(href=href, insert=img_loc, size=(insert_svg["im"].width, insert_svg["im"].height))
        dwg["svg"].add(image)
        
        if insert["framing"] is not None:
            # Calculate rx and ry as 5% of the width of the rectangle
            rx = ry = 0.05 * insert_svg["im"].width
        
            # Create the first rectangle with rounded corners
            rect1 = dwg["svg"].rect(
                insert=(img_loc[0], img_loc[1]),  # Bottom-left corner of the rectangle
                size=(insert_svg["im"].width, insert_svg["im"].height),  # Width and height of the rectangle
                fill="none",  # Fill color of the background (you can choose any color)
                stroke=mcolors.to_hex(insert["frame_color"]),  # Optional stroke for the rectangle
                stroke_width=insert["stroke_width"] * dwg["im"].width / 1000,
                rx=rx,
                ry=ry
            )
            dwg["svg"].add(rect1)
        
            # Create the second rectangle with rounded corners
            rect2 = dwg["svg"].rect(
                insert=(insert["framing"][0] - insert["framing"][2] / 2, insert["framing"][1] + insert["framing"][3] / 2),  # Bottom-left corner of the rectangle
                size=(insert["framing"][2], insert["framing"][3]),  # Width and height of the rectangle
                fill="none",  # Fill color of the background (you can choose any color)
                stroke=mcolors.to_hex(insert["frame_color"]),  # Optional stroke for the rectangle
                stroke_width=insert["stroke_width"] * dwg["im"].width / 1000,
                rx=rx,
                ry=ry
            )
            dwg["svg"].add(rect2)
        
            # Create a clipping path using the first rectangle
            clip_path = dwg["svg"].defs.add(dwg["svg"].clipPath(id="clip"))
            clip_path.add(dwg["svg"].rect(
                insert=(img_loc[0], img_loc[1]),
                size=(insert_svg["im"].width, insert_svg["im"].height),
                rx=rx,
                ry=ry
            ))
        
            # Apply the clipping path to the image
            image['clip-path'] = f'url(#{clip_path.get_id()})'
    # Save the SVG file
    dwg["svg"].save()
    
    
    return(dwg)
    
"""    
scalebar_style = {"frame":True,
                  "framepad":[30,2],
                  "stroke_width":4,
                  "stroke_style":"line",
                  "bar_color":"white",
                  "frame_color":"black",
                  "frame_opacity":0.6,
                  "location":"lower right",
                  "location_padding":[0.03,0.05],
                  "bar_ratio":[1/6,1/40]}

txt_style={"font":"Arial",
           "fontsize":34,
           "font_weight":"normal",
           "font_style":"normal",
           "text_decoration":"none",
           "color":"white"}
"""

def export_svg_to_png(inkscape_path, svg_file, output_png, width=None, height=None, dpi=None):
    """
    Export SVG to PNG using Inkscape command-line interface (CLI).
    
    Parameters:
    - inkscape_path: Path to the Inkscape executable
    - svg_file: Path to the input SVG file
    - output_png: Path to the output PNG file
    - width: Optional, desired width of the PNG
    - height: Optional, desired height of the PNG
    - dpi: Optional, desired DPI of the PNG
    """
    cmd = [inkscape_path, svg_file, '--export-type=png', '--export-filename=' + output_png]

    if width:
        cmd += ['--export-width=' + str(width)]
    if height:
        cmd += ['--export-height=' + str(height)]
    if dpi:
        cmd += ['--export-dpi=' + str(dpi)]

    subprocess.run(cmd)

